﻿using static System.Runtime.InteropServices.JavaScript.JSType;

class A {
	public void ProcessList(List<int> list) {
		// ... do something useful ...
		int index = list.FindIndex(IsGreaterThan5);

		Console.WriteLine($"A: found [{index}] == {list[index]}");
		// --- do more useful stuff ...
	}

	private static bool IsGreaterThan5(int x) => x > 5;
}

class B1 {
	public void ProcessList(List<int> list) {
		// ... do something useful ...
		int index = list.FindIndex(static (int x) => { return x > 5; });
		
		Console.WriteLine($"B1: found [{index}] == {list[index]}");
		// --- do more useful stuff ...
	}
}

class B2 {
	public void ProcessList(List<int> list) {
		// ... do something useful ...
		int index = list.FindIndex(static x => x > 5);

		Console.WriteLine($"B2: found [{index}] == {list[index]}");
		// --- do more useful stuff ...
	}
}

internal class Program {
	static void Main(string[] args) {
		var list = new List<int> { 2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31 };
		new A().ProcessList(list);
		new B1().ProcessList(list);
		new B2().ProcessList(list);

        Console.WriteLine();
        new C { TargetValue = 10 }.ProcessList(list);

        Console.WriteLine();
        new D().ProcessList(list);

		Console.WriteLine();
		new E().ProcessList(list);

		Console.WriteLine();
		new F().ProcessList(list);

		Console.WriteLine();
		new G().ProcessList(list);
	}
}

class C {
	public int TargetValue { get; set; }

	public void ProcessList(List<int> list) {
		int index = list.FindIndex(IsGreaterThanTargetValue);

		Console.WriteLine($"C: found [{index}] == {list[index]}");
	}

	private bool IsGreaterThanTargetValue(int x) => x > TargetValue;
}

class D {

	public void ProcessList(List<int> list) {
		int targetValue = 20;
        Console.WriteLine($"targetValue = {targetValue}");

		// C++ like capture by value (not available in C#):
		// int index = list.FindIndex(
		//   [=](int x) { return x > targetValue; }
		// );

		var scope = new Scope();
		scope.TargetValue = targetValue;
		int index = list.FindIndex(scope.GeneratedClosure);

		Console.WriteLine($"D: found [{index}] == {list[index]}");

		targetValue = 30;
		Console.WriteLine($"targetValue = {targetValue}");
	}

	private class Scope {
		public int TargetValue { get; set; }

		public bool GeneratedClosure(int x) => x > TargetValue;
	}
	
}

class E {

	public void ProcessList(List<int> list) {
		int targetValue = 20;

		Console.WriteLine($"targetValue = {targetValue}");

		// C++ like capture by reference (not available in C#):
		// int index = list.FindIndex(
		//   [&](int x) { return x > targetValue; }
		// );

		var scope = new Scope();
		// Error: Cannot ref-assign 'targetValue' to 'TargetValue' because 'targetValue' has a narrower escape scope than 'TargetValue'
		// scope.TargetValue = ref targetValue;
		int index = list.FindIndex(scope.GeneratedClosure);

		Console.WriteLine($"E: found [{index}] == {list[index]}");

		targetValue = 30;
		Console.WriteLine($"targetValue = {targetValue}");
	}

	private class Scope {
		// Error: A ref field can only be declared in a ref struct.
		// public ref int TargetValue;

		public bool GeneratedClosure(int x) => x > 0; // x > TargetValue;
	}

}

class F {

	public void ProcessList(List<int> list) {
		int targetValue = 20;

		Console.WriteLine($"targetValue = {targetValue}");

		// C# capture by moving - NOTE: same code as code in class G below:
		// int index = list.FindIndex(
		//   (int x) => { return x > targetValue; }
		// );

		var scope = new Scope();
		scope.TargetValue = targetValue;
		int index = list.FindIndex(scope.GeneratedClosure);

		Console.WriteLine($"F: found [{index}] == {list[index]}");

		scope.TargetValue = 30;										// Rewritten by C# compiler to target scope
		Console.WriteLine($"targetValue = {scope.TargetValue}");    // Rewritten by C# compiler to target scope
	}

	private class Scope {
		public int TargetValue { get; set; }

		public bool GeneratedClosure(int x) => x > TargetValue;
	}

}


class G {

	public void ProcessList(List<int> list) {
		int targetValue = 20;

		Console.WriteLine($"targetValue = {targetValue}");

		// C# capture by moving:
		int index = list.FindIndex(
			(int x) => { return x > targetValue; }
		);

		Console.WriteLine($"G: found [{index}] == {list[index]}");

		targetValue = 30;                                     
		Console.WriteLine($"targetValue = {targetValue}");    
	}
}